# S2MLCore version 1.0.0
# S2MLChecker.py
# by Antoine Rauzy
# Copyrights (c) 2018 AltaRica Association
#
# This file is part of S2MLCore.
# S2MLCore is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# S2MLCore is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with S2MLCore.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Creator
# 3) Undefined references
# 4) Circular definitions

# 1) Imported modules
# -------------------

import string
import sys
import re
import math

from S2MLModels import *

# 2) Creator
# ----------

class S2MLChecker:
  def __init__(self, engine):
    from S2MLEngine import S2MLEngine
    self._engine = engine
    self._model = None
    self._checkedElement = None
    self._checkedBlock = None
    self._currentBlock = None
    self._stack = []
    self._count = 0
    self._errors = 0

  def _RaiseError(self, message):
    self._errors += 1
    self._engine.RaiseError(message)

  def CheckModel(self, model):
    self.CheckReferences(model)
    if self._errors!=0:
      return self._errors
    self.CheckCircularDefinitions(model)
    return self._errors

# 3) Undefined references
# -----------------------

  def CheckReferences(self, model):
    self._model = model
    self._errors = 0
    self.CheckReferencesInBlock(self._model)
    return self._errors

  def CheckReferencesInBlock(self, block):
    for element in block.GetElements():
      if element.GetType()==S2MLType.BLOCK:
        previousCheckedBlock = self._checkedBlock
        self._checkedBlock = element
        self.CheckReferencesInBlock(element)
        self._checkedBlock = previousCheckedBlock
      elif element.GetType()==S2MLType.PARAMETER:
        self._checkedElement = element
        self._CheckReferencesInStochasticExpression(element.GetDefinition())
        self._checkedElement = None
    return self._errors

  def _CheckReferencesInAttributesOfElement(self, element):
    for name in element.GetAttributeNames():
      value = element.GetAttribute(name)
      self._CheckReferencesInStochasticExpression(value)

  def _CheckReferencesInStochasticExpression(self, expression):
    if expression.GetType()==S2MLType.REFERENCE_EXPRESSION:
      referencePath = expression.GetReferencePath()
      element = self._checkedBlock.AccessNamedElement(referencePath)
      if element==None:
        referenceIdentifier = referencePath.ToString()
        element = self._checkedBlock.LookForIdentifiedElement(referenceIdentifier)
      if element==None:
        checkedElementPath = self._checkedElement.GetPath()
        self._RaiseError("In block %s, undefined reference %s in definition of %s" % (self._checkedBlock.GetIdentifier(), referencePath.ToString(), checkedElementPath.ToString()))
        return
      if not self._model.IsValidReferredElementInStochasticExpression(element):
        checkedElementPath = self._checkedElement.GetPath()
        if checkedElementPath==None:
          self._RaiseError("In block %s, incorrect reference %s" % (self._checkedBlock.GetIdentifier(), referencePath.ToString()))
        else:
          self._RaiseError("In block %s, incorrect reference %s in definition of %s" % (self._checkedBlock.GetIdentifier(), referencePath.ToString(), checkedElementPath.ToString()))
        return
      expression.SetReferredElement(element)
    elif expression.GetType()==S2MLType.CONSTANT_EXPRESSION:
      pass
    else:
      for child in expression.GetChildren():
        self._CheckReferencesInStochasticExpression(child)

# 4) Circular definitions
# -----------------------

  def CheckCircularDefinitions(self, model):
    self._model = model
    self._errors = 0
    self.CheckCircularDefinitionsInBlock(self._model)
    self._model.ResetFlagsOfElements()
    return self._errors

  def CheckCircularDefinitionsInBlock(self, block):
    previousCheckedBlock = self._checkedBlock
    self._checkedBlock = block
    previousBlock = self._currentBlock
    self._currentBlock = block
    for element in block.GetElements():
      if element.GetFlag()==0:
        if element.GetType()==S2MLType.BLOCK:
          self.CheckCircularDefinitionsInBlock(element)
        if element.GetType()==S2MLType.PARAMETER:
          self.CheckCircularDefinitionsInDefinition(element)
    self._checkedBlock = previousCheckedBlock
    self._currentBlock = previousBlock
    return self._errors

  def CheckCircularDefinitionsInDefinition(self, element):
    if element.GetFlag()==2:
      return
    if element.GetFlag()==1:
      message = "In block %s, circular definition of %s\n" % (self._checkedBlock.GetAbsolutePath().ToString(), element.GetPath().ToString())
      for involvedParameter in self._stack:
        message += involvedParameter.GetIdentifier() + "\n"
      self._RaiseError(message)
      return
    element.SetFlag(1)
    self._stack.append(element)
    self.CheckCircularDefinitionsInStochasticExpression(element.GetDefinition())
    self._stack.pop()
    element.SetFlag(2)

  def CheckCircularDefinitionsInStochasticExpression(self, expression):
    if expression.GetType()==S2MLType.REFERENCE_EXPRESSION:
      referencePath = expression.GetReferencePath()
      element = self._currentBlock.AccessNamedElement(referencePath)
      if element==None:
        element = self._currentBlock.LookForIdentifiedElement(referencePath.ToString())
      if element==None:
        self.RaiseError("Internal error %s %s" % (self._currentBlock.GetAbsolutePath().ToString(), referencePath.ToString()))
        return
      if element.GetType()==S2MLType.PARAMETER:
        previousBlock = self._currentBlock
        self._currentBlock = element.GetContainer()
        self.CheckCircularDefinitionsInDefinition(element)
        self._currentBlock = previousBlock
    elif expression.GetType()==S2MLType.CONSTANT_EXPRESSION:
      pass
    else:
      for child in expression.GetChildren():
        self.CheckCircularDefinitionsInStochasticExpression(child)
