# S2MLCore version 1.0.0
# S2MLStochasticExpressions.py
# by Antoine Rauzy
# Copyrights (c) 2018 AltaRica Association
#
# This file is part of S2MLCore.
# S2MLCore is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# S2MLCore is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with S2MLCore.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Stochastic expressions
# 3) Manager

# 1) Imported modules
# -------------------

import string
import sys
import re
import math

from S2MLType import *
from S2MLMetaModel import *
from S2MLConstants import *

# 2) Parameters
# -------------

class S2MLParameter(S2MLElement):
  def __init__(self, path, container):
    S2MLElement.__init__(self, S2MLType.PARAMETER, container)
    self.SetPath(path)
    self._definition = None
    self._value = S2MLConstant.NewNone()

  def GetDefinition(self):
    return self._definition
  def SetDefinition(self, expression):
    self._definition = expression

  def IsValueCalculated(self):
    return not self._value.IsNone()
  def ResetValue(self):
    self._value.Reset()
  def GetValue(self):
    return self._value
  def SetValue(self, value):
    self._value.Set(value)


# 2) Stochastic expressions
# -------------------------

class S2MLStochasticExpression(S2MLTerm):

  def __init__(self, type):
    S2MLTerm.__init__(self, type)
    self._value = S2MLConstant.NewNone()

  @staticmethod
  def NewReference(path):
    expression = S2MLStochasticExpression(S2MLType.REFERENCE_EXPRESSION)
    expression.SetReferencePath(path)
    return expression

  @staticmethod
  def NewConstant(value):
    expression = S2MLStochasticExpression(S2MLType.CONSTANT_EXPRESSION)
    expression.SetValue(value)
    return expression

  @staticmethod
  def NewOperation(type, children):
    expression = S2MLStochasticExpression(type)
    expression.AppendChildren(children)
    return expression

  def Copy(self):
    if self.GetType()==S2MLType.REFERENCE_EXPRESSION:
      myPath = self.GetReferencePath()
      path = S2MLPath.Copy(myPath)
      return S2MLStochasticExpression.NewReference(path)
    if self.GetType()==S2MLType.CONSTANT_EXPRESSION:
      myValue = self.GetValue()
      value = S2MLConstant.Copy(myValue)
      return S2MLStochasticExpression.NewConstant(value)
    children = []
    for myChild in self.GetChildren():
      child = myChild.Copy()
      children.append(child)
    return S2MLStochasticExpression.NewOperation(self.GetType(), children)

  def IsValueCalculated(self):
    return not self._value.IsNone()
  def ResetValue(self):
    self._value.Reset()
  def GetValue(self):
    return self._value
  def SetValue(self, value):
    self._value.Set(value)

# 3) Operators
# ------------

class S2MLStochasticOperator:
  MANY = -1
  def __init__(self, type, keywordCode, minimumNumberOfArguments, maximumNumberOfArguments, calculate):
    self._type = type
    self._keywordCode = keywordCode
    self._minimumNumberOfArguments = minimumNumberOfArguments
    self._maximumNumberOfArguments = maximumNumberOfArguments
    self._calculate = calculate

  def GetType(self):
    return self._type

  def GetKeywordCode(self):
    return self._keywordCode

  def GetMinimumNumberOfArguments(self):
    return self._minimumNumberOfArguments

  def GetMaximumNumberOfArguments(self):
    return self._maximumNumberOfArguments

  def GetCalculate(self):
    return self._calculate

# 4) Manager
# ----------


class S2MLStochasticExpressionManager:

# 4.1) Creator
# ------------

  def __init__(self, engine):
    from S2MLEngine import S2MLEngine
    self._engine = engine
    self._operatorsByTypes = dict()
    self._operatorsByKeywordCodes = dict()
    self.RegisterOperators()
    self._missionTime = 0.0

  def RegisterOperators(self):
    self.RegisterOperator(S2MLType.ADD_EXPRESSION, S2MLType.KEYWORD_ADD,   1, S2MLStochasticOperator.MANY, S2MLStochasticExpressionManager.CalculateAdd)
    self.RegisterOperator(S2MLType.SUB_EXPRESSION, S2MLType.KEYWORD_SUB,   1, S2MLStochasticOperator.MANY, S2MLStochasticExpressionManager.CalculateSub)
    self.RegisterOperator(S2MLType.MUL_EXPRESSION, S2MLType.KEYWORD_MUL,   1, S2MLStochasticOperator.MANY, S2MLStochasticExpressionManager.CalculateMul)
    self.RegisterOperator(S2MLType.DIV_EXPRESSION, S2MLType.KEYWORD_DIV,   1, S2MLStochasticOperator.MANY, S2MLStochasticExpressionManager.CalculateDiv)
    self.RegisterOperator(S2MLType.NEG_EXPRESSION, S2MLType.KEYWORD_NEG,   1, 1,                           S2MLStochasticExpressionManager.CalculateNeg)
    self.RegisterOperator(S2MLType.MIN_EXPRESSION, S2MLType.KEYWORD_MIN,   1, S2MLStochasticOperator.MANY, S2MLStochasticExpressionManager.CalculateMin)
    self.RegisterOperator(S2MLType.MAX_EXPRESSION, S2MLType.KEYWORD_MAX,   1, S2MLStochasticOperator.MANY, S2MLStochasticExpressionManager.CalculateMax)
    self.RegisterOperator(S2MLType.AND_EXPRESSION, S2MLType.KEYWORD_AND,   1, S2MLStochasticOperator.MANY, S2MLStochasticExpressionManager.CalculateAnd)
    self.RegisterOperator(S2MLType.OR_EXPRESSION,  S2MLType.KEYWORD_OR,    1, S2MLStochasticOperator.MANY, S2MLStochasticExpressionManager.CalculateOr)
    self.RegisterOperator(S2MLType.NOT_EXPRESSION, S2MLType.KEYWORD_NOT,   1, 1,                           S2MLStochasticExpressionManager.CalculateNot)
    self.RegisterOperator(S2MLType.EQ_EXPRESSION,  S2MLType.KEYWORD_EQ,    2, 2,                           S2MLStochasticExpressionManager.CalculateEq)
    self.RegisterOperator(S2MLType.DF_EXPRESSION,  S2MLType.KEYWORD_DF,    2, 2,                           S2MLStochasticExpressionManager.CalculateDf)
    self.RegisterOperator(S2MLType.LT_EXPRESSION,  S2MLType.KEYWORD_LT,    2, 2,                           S2MLStochasticExpressionManager.CalculateLt)
    self.RegisterOperator(S2MLType.GT_EXPRESSION,  S2MLType.KEYWORD_GT,    2, 2,                           S2MLStochasticExpressionManager.CalculateGt)
    self.RegisterOperator(S2MLType.LEQ_EXPRESSION, S2MLType.KEYWORD_LEQ,   2, 2,                           S2MLStochasticExpressionManager.CalculateLeq)
    self.RegisterOperator(S2MLType.GEQ_EXPRESSION, S2MLType.KEYWORD_GEQ,   2, 2,                           S2MLStochasticExpressionManager.CalculateGeq)
    self.RegisterOperator(S2MLType.GEQ_EXPRESSION, S2MLType.KEYWORD_COUNT, 1, S2MLStochasticOperator.MANY, S2MLStochasticExpressionManager.CalculateCount)
    self.RegisterOperator(S2MLType.GEQ_EXPRESSION, S2MLType.KEYWORD_IF,    3, 3,                           S2MLStochasticExpressionManager.CalculateIf)

  def GetEngine(self):
    return self._engine

# 4.2) Management of operators
# ----------------------------

  def GetOperatorOfType(self, type):
    return self._operatorsByTypes.get(type, None)

  def GetOperatorOfKeywordCode(self, keywordCode):
    return self._operatorsByKeywordCodes.get(keywordCode, None)

  def RegisterOperator(self, type, keywordCode, minimumNumberOfArguments, maximumNumberOfArguments, calculate):
    if self.GetOperatorOfType(type)!=None or self.GetOperatorOfKeywordCode(keywordCode)!=None:
      return
    operator = S2MLStochasticOperator(type, keywordCode, minimumNumberOfArguments, maximumNumberOfArguments, calculate)
    self._operatorsByTypes[type] = operator
    self._operatorsByKeywordCodes[keywordCode] = operator

  def GetKeywordCodeOfOperatorOfType(self, type):
    operator = self.GetOperatorOfType(type)
    if operator==None:
      return 0
    return operator._keywordCode

# 4.3) Calculation of values of expresssions
# ------------------------------------------

  def ResetValuesOfParameters(self, container):
    for element in container.GetElements():
      if element.GetType()==S2MLType.BLOCK:
        self.ResetValuesOfParameters(element)
      elif element.GetType()==S2MLType.PARAMETER:
        element.ResetValue()

  def CalculateValueOfParameter(self, parameter):
    if parameter.IsValueCalculated():
      value = parameter.GetValue()
    else:
      definition = parameter.GetDefinition()
      value = self.CalculateValueOfExpression(definition)
      parameter.SetValue(value)
    # print("parameter(" + parameter.GetIdentifier() + ") = " + str(value.ToFloat()))
    return value

  def CalculateValueOfExpression(self, expression):
    if expression.GetType()==S2MLType.REFERENCE_EXPRESSION:
      referredElement = expression.GetReferredElement()
      value = self.CalculateValueOfParameter(referredElement)
      expression.SetValue(value)
      return expression.GetValue()
    if expression.GetType()==S2MLType.CONSTANT_EXPRESSION:
      return expression.GetValue()
    operator = self.GetOperatorOfType(expression.GetType())
    if operator==None:
      return S2MLConstant.NewNone()
    return operator._calculate(self, expression)

  def CalculateAdd(self, expression):
    value = expression.GetValue()
    value.SetAsInteger(0)
    for child in expression.GetChildren():
      childValue = self.CalculateValueOfExpression(child)
      value.AddMe(childValue)
    return value

  def CalculateSub(self, expression):
    value = expression.GetValue()
    value.SetAsInteger(0)
    first = True
    for child in expression.GetChildren():
      childValue = self.CalculateValueOfExpression(child)
      if first:
        first = False
        value.Set(childValue)
      else:
        value.SubMe(childValue)
    return value

  def CalculateMul(self, expression):
    value = expression.GetValue()
    value.SetAsInteger(1)
    for child in expression.GetChildren():
      childValue = self.CalculateValueOfExpression(child)
      value.MulMe(childValue)
    return value

  def CalculateDiv(self, expression):
    value = expression.GetValue()
    value.SetAsInteger(0)
    first = True
    for child in expression.GetChildren():
      childValue = self.CalculateValueOfExpression(child)
      if first:
        first = False
        value.Set(childValue)
      else:
        value.DivMe(childValue)
    return value

  def CalculateNeg(self, expression):
    value = expression.GetValue()
    child = expression.GetChild(0)
    childValue = self.CalculateValueOfExpression(child)
    value.NegMe(childValue)
    return value

  def CalculateMin(self, expression):
    value = expression.GetValue()
    value.SetAsPosInf()
    first = True
    for child in expression.GetChildren():
      childValue = self.CalculateValueOfExpression(child)
      value.MinMe(childValue)
    return value

  def CalculateMax(self, expression):
    value = expression.GetValue()
    value.SetAsNegInf()
    first = True
    for child in expression.GetChildren():
      childValue = self.CalculateValueOfExpression(child)
      value.MaxMe(childValue)
    return value

  def CalculateAnd(self, expression):
    value = expression.GetValue()
    value.SetAsBoolean(True)
    for child in expression.GetChildren():
      childValue = self.CalculateValueOfExpression(child)
      value.AndMe(childValue)
    return value

  def CalculateOr(self, expression):
    value = expression.GetValue()
    value.SetAsBoolean(False)
    for child in expression.GetChildren():
      childValue = self.CalculateValueOfExpression(child)
      value.OrMe(childValue)
    return value

  def CalculateNot(self, expression):
    value = expression.GetValue()
    value.SetAsBoolean(False)
    child = expression.GetChild(0)
    childValue = self.CalculateValueOfExpression(child)
    value.NotMe(childValue)
    return value

  def CalculateEq(self, expression):
    value = expression.GetValue()
    value.SetAsBoolean(False)
    child1 = expression.GetChild(0)
    childValue1 = self.CalculateValueOfExpression(child1)
    child2 = expression.GetChild(1)
    childValue2 = self.CalculateValueOfExpression(child2)
    value.EqMe(childValue1, childValue2)
    return value

  def CalculateDf(self, expression):
    value = expression.GetValue()
    value.SetAsBoolean(False)
    child1 = expression.GetChild(0)
    childValue1 = self.CalculateValueOfExpression(child1)
    child2 = expression.GetChild(1)
    childValue2 = self.CalculateValueOfExpression(child2)
    value.DfMe(childValue1, childValue2)
    return value

  def CalculateLt(self, expression):
    value = expression.GetValue()
    value.SetAsBoolean(False)
    child1 = expression.GetChild(0)
    childValue1 = self.CalculateValueOfExpression(child1)
    child2 = expression.GetChild(1)
    childValue2 = self.CalculateValueOfExpression(child2)
    value.LtMe(childValue1, childValue2)
    return value

  def CalculateGt(self, expression):
    value = expression.GetValue()
    value.SetAsBoolean(False)
    child1 = expression.GetChild(0)
    childValue1 = self.CalculateValueOfExpression(child1)
    child2 = expression.GetChild(1)
    childValue2 = self.CalculateValueOfExpression(child2)
    value.GtMe(childValue1, childValue2)
    return value

  def CalculateLeq(self, expression):
    value = expression.GetValue()
    value.SetAsBoolean(False)
    child1 = expression.GetChild(0)
    childValue1 = self.CalculateValueOfExpression(child1)
    child2 = expression.GetChild(1)
    childValue2 = self.CalculateValueOfExpression(child2)
    value.LeqMe(childValue1, childValue2)
    return value

  def CalculateGeq(self, expression):
    value = expression.GetValue()
    value.SetAsBoolean(False)
    child1 = expression.GetChild(0)
    childValue1 = self.CalculateValueOfExpression(child1)
    child2 = expression.GetChild(1)
    childValue2 = self.CalculateValueOfExpression(child2)
    value.GeqMe(childValue1, childValue2)
    return value

  def CalculateCount(self, expression):
    value = expression.GetValue()
    value.SetAsInteger(0)
    for child in expression.GetChildren():
      childValue = self.CalculateValueOfExpression(child)
      value.CountMe(childValue)
    return value

  def CalculateIf(self, expression):
    value = expression.GetValue()
    child1 = expression.GetChild(0)
    childValue1 = self.CalculateValueOfExpression(child1)
    if childValue.IsBoolean():
      if childValue.GetValue():
        child2 = expression.GetChild(1)
      else:
        child2 = expression.GetChild(2)
      childValue2 = self.CalculateValueOfExpression(child2)
      value.Set(childValue2)
    else:
      value.Reset()
